﻿using System.Collections;
using System.Collections.Generic;
using System;
using UnityEngine;

namespace Assets.Scripts.Systems
{
    public class EnemySpawnMaster
    {
        private Dictionary<HexSide, SpawnSide> spawnSides;
        private static System.Random random;
        public EnemySpawnMaster(int gridEdgeSize, float maxIntensity)
        {
            random = new System.Random();
            spawnSides = new Dictionary<HexSide, SpawnSide>();
            foreach(HexSide side in Enum.GetValues(typeof(HexSide)))
            {
                spawnSides.Add(side, new SpawnSide(gridEdgeSize, side, maxIntensity));
            }
        }

        public void SetSideIntensities(HexSide side, float intensity)
        {
            SpawnSide spawnSide;
            if (spawnSides.TryGetValue(side, out spawnSide))
            {
                spawnSide.UpdateIntensities(intensity);
            }
        }

        public void CallSpawns(SpawnCallback callback, float spawnThreshold)
        {
            foreach(SpawnSide side in spawnSides.Values)
            {
                side.CallSpawns(callback, spawnThreshold);
            }
        }

        private class SpawnSide
        {
            private int gridEdgeSize;
            private HexSide side;
            private List<Spawner> spawns;
            float maxIntensity;
            public SpawnSide(int gridEdgeSize, HexSide side, float maxIntensity)
            {
                this.gridEdgeSize = gridEdgeSize;
                this.side = side;
                this.maxIntensity = maxIntensity;
                int halfGrid = gridEdgeSize / 2;
                switch(side)
                {
                    case HexSide.Top:
                        spawns = CreateSpawns(new Vector2(halfGrid, -gridEdgeSize), new Vector2(1, 0));
                        break;
                    case HexSide.TopLeft:
                        spawns = CreateSpawns(new Vector2(-halfGrid, -halfGrid), new Vector2(1, -1));
                        break;
                    case HexSide.TopRight:
                        spawns = CreateSpawns(new Vector2(gridEdgeSize, -halfGrid), new Vector2(0, 1));
                        break;
                    case HexSide.Bottom:
                        spawns = CreateSpawns(new Vector2(-halfGrid, gridEdgeSize), new Vector2(-1, 0));
                        break;
                    case HexSide.BottomLeft:
                        spawns = CreateSpawns(new Vector2(-gridEdgeSize, halfGrid), new Vector2(0, -1));
                        break;
                    case HexSide.BottomRight:
                        spawns = CreateSpawns(new Vector2(halfGrid, halfGrid), new Vector2(-1, 1));
                        break;
                }
            }

            // Intensity drops the further away from center
            public void UpdateIntensities(float centerIntensity)
            {
                foreach(Spawner spawn in spawns)
                {
                    spawn.SetIntensity(centerIntensity > maxIntensity ? maxIntensity : centerIntensity);
                    centerIntensity -= .1f;
                }
            }

            public void CallSpawns(SpawnCallback callback, float spawnThreshold)
            {
                foreach(Spawner spawner in spawns)
                {
                    spawner.AskForSpawn(callback, spawnThreshold);
                }
            }

            private List<Spawner> CreateSpawns(Vector2 startLocation, Vector2 oddOffset)
            {
                List<Spawner> spawns = new List<Spawner>();
                spawns.Add(new Spawner(startLocation));
                switch(side)
                {
                    case HexSide.Top:
                    case HexSide.BottomLeft:
                        spawns.Add(new Spawner(startLocation + (oddOffset / 2)));
                        break;
                    case HexSide.TopRight:
                    case HexSide.Bottom:
                    case HexSide.TopLeft:
                    case HexSide.BottomRight:
                        spawns.Add(new Spawner(startLocation - (oddOffset / 2)));
                        break;
                }
                for (int i = 2; i < gridEdgeSize - 1; i++)
                {
                    float distance = i % 2 == 0 ? Mathf.Floor(-i / 2) : Mathf.Floor(i / 2);
                    spawns.Add(new Spawner(startLocation + (distance * oddOffset)));
                }
                return spawns;
            }
        }

        private class Spawner
        {
            // intensity of 0 means it will spawn nothing
            private Vector2 location;
            private float intensity;
            public Spawner(Vector2 location)
            {
                this.location = location;
                intensity = 0;
                Debug.Log("Created spawner at " + location.ToString());
            }

            public void SetIntensity(float intensity)
            {
                this.intensity = intensity;
            }

            public void AskForSpawn(SpawnCallback callback, float spawnThreshold)
            {
                if (random.NextDouble() * intensity > spawnThreshold)
                {
                    callback(location, (float)random.NextDouble() * intensity);
                }
            }
        }
    }

    public enum HexSide { Top, TopLeft, TopRight, Bottom, BottomLeft, BottomRight }

    public delegate void SpawnCallback(Vector2 location, float dangerFactor);
}