﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace Assets.Scripts.Systems
{
    class GameManager : MonoBehaviour
    {
        private const float ATTACK_INTERVAL = 2f;
        private const float SPAWN_THRESHOLD = .9f;
        private const int GRID_RADIUS = 10;
        public const float MAX_INTENSITY = 1.3f;
        private const float ENEMY_HEIGHT = 1f;

        private static readonly float X_STEP = Mathf.Sqrt(1 - Mathf.Pow(.5f, 2));

        public static GameManager instance;

        private GameState currentState;
        private Turret currentTurret;
        private int wave;
        EnemySpawnMaster spawnMaster;
        private float timeSinceLastAttack;
        private float waveTime;

        private Vector3 center; // Center of map
        private float hexSize;  //Flat side to opposite flat side

        public const float MAX_POWER = 100;
        private float currPower;

        private float currency;

        private Dictionary<Hexagon, GameObject> occupiedHexagons;

        private AudioSource buildingMusic;
        private AudioSource battleMusic;

        private Material machineOn;
        private Material machineOff;

        private void Awake()
        {
            if (instance == null)
            {
                instance = this;
            }
            else if (instance != this)
            {
                Destroy(this);
            }

            DontDestroyOnLoad(this);

            InitGame();
        }

        private void InitGame()
        {
            machineOn = (Material)Resources.Load("MachineGunOnMaterial");
            machineOff = (Material)Resources.Load("MachineGunOffMaterial");
            Color machineColor = machineOff.color;
            machineColor.a = .5f;
            machineOff.color = machineColor;

            Application.targetFrameRate = 60;
            currentState = GameState.Building;
            wave = 4;
            spawnMaster = new EnemySpawnMaster(GRID_RADIUS, MAX_INTENSITY);
            timeSinceLastAttack = 0f;

            currPower = MAX_POWER;
            currency = 300  + (50 * wave);

            center = GameObject.FindGameObjectWithTag("CenterHexagon").GetComponent<Collider>().bounds.center;
            hexSize = GameObject.FindGameObjectWithTag("CenterHexagon").GetComponent<MyHexagon>().GetCellDimensions().x;

            occupiedHexagons = new Dictionary<Hexagon, GameObject>();

            currentTurret = Turret.MachineGun;

            buildingMusic = GameObject.Find("BuildingMusic").GetComponent<AudioSource>();
            battleMusic = GameObject.Find("BattleMusic").GetComponent<AudioSource>();

            buildingMusic.Play();
            //For testing
            //BeginBattle();
        }

        private void FixedUpdate()
        {
            switch (currentState)
            {
                case GameState.Battle:
                    waveTime -= Time.deltaTime;
                    if (waveTime > 0)
                    {
                        timeSinceLastAttack += Time.deltaTime;
                        if (timeSinceLastAttack > ATTACK_INTERVAL)
                        {
                            spawnMaster.CallSpawns(SpawnEnemy, SPAWN_THRESHOLD);
                            timeSinceLastAttack = 0;
                        }
                    }
                    else if (GameObject.FindGameObjectsWithTag("Enemy").Count() == 0)
                    {
                        currentState = GameState.Building;
                        battleMusic.Stop();
                        buildingMusic.Play();
                        wave++;
                    }
                    break;
            }
        }

        private void SpawnEnemy(Vector2 location, float dangerFactor)
        {
            Vector3 position = PositionFromLocation(location);
            
            if (dangerFactor > 3)
            {
                GameObject enemy = (GameObject)Instantiate(Resources.Load("Enemy"), position, Quaternion.identity);
                enemy.GetComponent<AIController>().type = EnemyType.demolition;
            }
            else
            {
                //GameObject enemy = (GameObject)Instantiate(Resources.Load("Enemy"), position, Quaternion.identity);
                GameObject enemy = (GameObject)Instantiate(Resources.Load("Skitter"), position, Quaternion.identity);
                enemy.GetComponent<AIController>().type = EnemyType.Skitter;
            }
        }

        private Vector3 PositionFromLocation(Vector2 location)
        {
            return center + new Vector3(location.x * hexSize * X_STEP, ENEMY_HEIGHT, (location.y * hexSize) + (location.x * hexSize * .5f));
        }

        public void BeginBattle()
        {
            buildingMusic.Stop();
            battleMusic.Play();
            timeSinceLastAttack = 0f;
            currentState = GameState.Battle;
            waveTime = Assets.Scripts.Utilities.Difficulty.Set(spawnMaster, wave);
        }

        public GameState GetState()
        {
            return currentState;
        }

        public bool DrainPower(float drainAmount)
        {
            if (currPower >= drainAmount)
            {
                currPower -= drainAmount;
                return true;
            }
            return false;
        }

        public void ReturnPower(float returnAmount)
        {
            currPower = Mathf.Min(currPower + returnAmount, MAX_POWER);
        }

        public float CurrentPower
        {
            get
            {
                return currPower;
            }
        }

        public float Currency
        {
            get
            {
                return currency;
            }
        }

        public bool SpendCurrency(float amount)
        {
            if (currency >= amount)
            {
                currency -= amount;
                return true;
            }
            return false;
        }

        public void AddCurrency(float amount)
        {
            currency += amount;
        }

        public int GetWave()
        {
            return wave;
        }

        public Dictionary<Hexagon, GameObject> OccupiedHexagons
        {
            get
            {
                return occupiedHexagons;
            }
        }

        public bool BuyTurret()
        {
            switch(currentTurret)
            {
                case Turret.MachineGun:
                    return SpendCurrency(100);
            }
            return false;
        }

        public void DeadEnemy(EnemyType enemy)
        {
            switch(enemy)
            {
                case EnemyType.Skitter:
                    AddCurrency(wave * 5);
                    break;
            }
        }

        public Material GetMachineMaterial(bool on)
        {
            return on ? machineOn : machineOff;
        }
    }

    public enum GameState { Building, Battle, Pause }

    public enum Turret { MachineGun };
}
